from thesoup.utilityclasses.graph import Graph
from thesoup.utilityclasses.disjointsets import DisjointSets


def _topological_explore(graph: Graph, visited: set, stk: list, start):
    if start not in visited:
        visited.add(start)
        for c, _ in graph.get_neighbours(start):
            _topological_explore(graph, visited, stk, c)
        stk.append(start)


def topological_sort(graph: Graph) -> list:
    """
    This function returns a lost of vertices in a topologically sorted order.
    :param graph: A directed graph.
    :return: A list of sorted vertices.
    """
    vertices = graph.vertices()
    visited = set()
    stk = list()
    for v in vertices:
        _topological_explore(graph, visited, stk, v)
    return reversed(stk)


def kruskal(graph: Graph) -> set:
    """
    This function implements the Kruskal's algorithm to find the min cost spanning tree of a di-graph
    :param graph: The input graph
    :return: A list of the edges comprising the MCST
    """
    sorted_edges = sorted(graph.edges())
    ds = DisjointSets(graph.vertices())
    selected_edges = set()

    for edge in sorted_edges:
        src, dest = edge.src, edge.destination
        if ds.find_set(src) != ds.find_set(dest):
            ds.union(src, dest)
        selected_edges.add(edge)
        if len(ds) == 1:
            break
    return selected_edges



class CountSet:
    """
    This class implements a set where if multiple elements are inserted, the set keeps track of counts.
    It supports the usual set functionality like add, update, etc. Along with that it also supports access with square
    brackets (__getitem__) that returns the count of an element. If absent, 0 is returned
    """
    def __init__(self, iterable=None):
        self.num = 0
        self.store = dict()
        if iterable is not None:
            for elem in iterable:
                self.add(elem)

    def _add_many(self, items: list):
        for item, count in items:
            if item in self.store:
                self.store[item] += count
            else:
                self.store[item] = count
            self.num += count

    def add(self, item):
        """
        Method to add an item to the set.

        :param item: The item to add
        :return: Void
        """
        if item not in self.store:
            self.store[item] = 1
        else:
            self.store[item] += 1
        self.num += 1

    def update(self, items):
        """
        Method to add many items to the set.

        :param items: The iterable of items to add
        :return: Void
        """
        for item in items:
            self.add(item)

    def __contains__(self, item) -> bool:
        return item in self.store

    def __getitem__(self, item) -> int:
        return 0 if item not in self.store else self.store[item]

    def __add__(self, other: 'CountSet') -> 'CountSet':
        if type(other) != CountSet:
            raise TypeError(f"Cannot add {type(other)} with CountSet object")
        added = CountSet()
        added.store = self.store.copy()
        added.num = self.num
        added._add_many(other.store.items())
        return added

    def __iter__(self):
        return iter(self.store.items())

    def __delitem__(self, key):
        if key in self.store:
            if self.store[key] == 1:
                del self.store[key]
            else:
                self.store[key] -= 1
            self.num -= 1

    def __eq__(self, other):
        return type(other) == CountSet and self.store == other.store

    def __len__(self):
        return self.num



def merge(*arrays: list) -> list:
    """
    This function merges K individually sorted lists into a single sorted list. Complexity is `KNlog(N)`, where we are
    merging K arrays of N size each.
    :params arrays: A vararg of lists.
    :return: A new merged list.
    """
    if len(arrays) == 1:
        return arrays[0]
    semi_merged_list = list()
    for i in range(0, len(arrays)-1, 2):
        it1 = 0
        it2 = 0
        semi_merged_list.append(list())
        while it1 < len(arrays[i]) and it2 < len(arrays[i+1]):
            if arrays[i][it1] < arrays[i+1][it2]:
                semi_merged_list[-1].append(arrays[i][it1])
                it1 += 1
            else:
                semi_merged_list[-1].append(arrays[i+1][it2])
                it2 += 1
        while it1 < len(arrays[i]):
            semi_merged_list[-1].append(arrays[i][it1])
            it1 += 1
        while it2 < len(arrays[i+1]):
            semi_merged_list[-1].append(arrays[i+1][it2])
            it2 += 1
    if len(arrays) % 2 == 1:
        semi_merged_list.append(arrays[-1])
    return merge(*semi_merged_list)



class DisjointSets:
    """
    This class implements the disjoint sets data structure.
    """
    def __init__(self, items: list):
        self.set_map = dict(
            [(item, item) for item in items]
        )
        self.sets = dict(
            [(item, {item}) for item in items]
        )

    def find_set(self, item):
        return self.set_map[item]

    def union(self, item1, item2):
        dest_set = self.set_map[item1]
        src_set = self.set_map[item2]
        if src_set != dest_set:
            for item in self.sets[src_set]:
                self.set_map[item] = self.set_map[dest_set]
                self.sets[dest_set].add(item)
            del self.sets[src_set]

    def __len__(self):
        return len(self.sets)


from setuptools import setup
from os import path
from io import open
import sys

here = path.abspath(path.dirname(__file__))

version_maj = sys.version_info[0]
version_min = sys.version_info[1]

if version_maj < 3:
    sys.exit("Need python 3 to run. You are trying to install this with python {}.{}".format(version_maj, version_min))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="thesoup",
    version="0.8",
    description="A soup of random python utils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/amartya00/thesoup",
    author="Amartya Datta Gupta",
    author_email="amartya00@gmail.com",
    license='GPLv2',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 3.9',
    ],
    packages=["thesoup.utilityclasses", "thesoup.utilityfunctions"],
    install_requires=[],
    extras_require={
        'dev': [],
        'test': ["nose"],
    },
    project_urls={
        "Source": "https://github.com/amartya00/thesoup"
    },
)



from abc import ABC, abstractmethod


class _BinaryTreeElement:
    def __init__(self, val):
        self.val = val
        self.count = 1
        self.right = None
        self.left = None
        self.parent = None

    def __call__(self, *args, **kwargs):
        return self.val


class BinaryTree (ABC):
    """
    This is the interface for a binary tree. Learn more about binary trees here: https://en.wikipedia.org/wiki/Binary_tree.
    Any implementation of this should provide methods for:
    1. Insert
    2. Search
    3. Delete
    4. Size
    5. Iterator
    """
    @abstractmethod
    def __contains__(self, item):
        """
        Method to check for an element in the tree

        :param item: The item to search
        :return: A boolean indicating presence
        """
        pass

    @abstractmethod
    def insert(self, item):
        """
        Method to insert an element in the tree

        :param item: The item to insert.
        :return:
        """
        pass

    @abstractmethod
    def __delitem__(self, key):
        """
        Method to delete an item from the tree

        :param item: The item to search and delete
        """
        pass

    @abstractmethod
    def __len__(self):
        """
        Method returning the size of the tree

        :return: The number of elements in the tree
        """
        pass

    @abstractmethod
    def __iter__(self):
        """
        Provide an iterator for a binary tree
        """
        pass


class BinarySearchTree (BinaryTree):
    """
    A BST implementation of a binary tree
    """
    def __init__(self):
        self.root = None
        self.num = 0

    @staticmethod
    def _insert(root: _BinaryTreeElement, element):
        if element > root():
            if root.right is not None:
                BinarySearchTree._insert(root.right, element)
            else:
                root.right = _BinaryTreeElement(element)
                root.right.parent = root
        elif element < root():
            if root.left is not None:
                BinarySearchTree._insert(root.left, element)
            else:
                root.left = _BinaryTreeElement(element)
                root.left.parent = root
        else:
            root.count += 1

    @staticmethod
    def _search(root: _BinaryTreeElement, item):
        iterator = root
        while iterator is not None:
            if iterator() == item:
                return iterator
            elif item > iterator():
                iterator = iterator.right
            else:
                iterator = iterator.left
        return None

    @staticmethod
    def _min(root: _BinaryTreeElement):
        iterator = root
        while iterator.left is not None:
            iterator = iterator.left
        return iterator

    @staticmethod
    def _max(root: _BinaryTreeElement):
        iterator = root
        while iterator.right is not None:
            iterator = iterator.right
        return iterator

    @staticmethod
    def _hard_delete(item: _BinaryTreeElement):
        if item.left is not None:
            max_left = BinarySearchTree._max(item.left)
            item.val = max_left.val
            item.count = max_left.count
            BinarySearchTree._hard_delete(max_left)
        elif item.right is not None:
            min_right = BinarySearchTree._min(item.right)
            item.val = min_right.val
            item.count = min_right.count
            BinarySearchTree._hard_delete(min_right)
        else:
            if item() < item.parent():
                item.parent.left = None
                del item
            else:
                item.parent.right = None
                del item

    @staticmethod
    def _next(root: _BinaryTreeElement):
        if root.right is not None:
            return BinarySearchTree._min(root.right)
        elif root.parent is not None:
            if root.parent() > root():
                return root.parent
            else:
                while root.parent is not None and root.parent() < root():
                    root = root.parent
                return root.parent
        else:
            return None

    def insert(self, item):
        if self.root is None:
            self.root = _BinaryTreeElement(item)
        else:
            BinarySearchTree._insert(self.root, item)
        self.num += 1

    def __len__(self):
        return self.num

    def __contains__(self, item):
        return BinarySearchTree._search(self.root, item) is not None

    def __delitem__(self, key):
        # Special case for root
        if key == self.root() and len(self) == 1:
            tmp = self.root
            self.root = None
            del tmp
            self.num = 0
        item = BinarySearchTree._search(self.root, key)
        if item is not None:
            BinarySearchTree._hard_delete(item)
            self.num -= 1

    def __iter__(self):
        class _BinarySearchTreeIterator:
            def __init__(self, root):
                self.pos = root
                self.pos_count = root.count

            def __next__(self):
                if self.pos is None:
                    raise StopIteration()
                elif self.pos_count > 1:
                    self.pos_count -= 1
                    return self.pos()
                else:
                    tmp = self.pos
                    self.pos = BinarySearchTree._next(self.pos)
                    if self.pos is not None:
                        self.pos_count = self.pos.count
                    return tmp()

        return _BinarySearchTreeIterator(
            BinarySearchTree._min(
                self.root
            )
        )